@@ -7,6 +7,7 @@ use std::fs::File;
 use std::num::NonZeroUsize;
 use unicode_segmentation::UnicodeSegmentation;
 
+
 /* Byte formatting stuff lifted from hexyl */
 pub enum ByteCategory {
     Null,
@@ -97,7 +98,7 @@ pub struct State {
     pub show_prompt: bool,
     pub color: bool,
 
-    /* Current byte number, 0 to len -1 */
+    /* Current byte number, 0 to (len - 1) */
     pub index: usize,
 
     pub width: NonZeroUsize,
@@ -108,6 +109,12 @@ pub struct State {
     /* Spaces to put between a byte number and a byte when displaying */
     pub n_padding: String,
 
+    /* Number of lines to print before current line */
+    pub before_context: usize,
+
+    /* Number of lines to print after current line */
+    pub after_context: usize,
+
     pub last_search: Option<Vec<u8>>,
 }
 
@@ -170,8 +177,76 @@ impl State {
         to_return
     }
 
+
+    /// Return the range `self.width` bytes starting at `address`.
+    /// Could be cut short by hitting end of all bytes.
+    /// Could be empty because `address` is past end of all bytes.
+    pub fn bytes_range_from(&self, address:usize) -> std::ops::Range<usize> {
+        let width = usize::from(self.width);
+        if address < self.all_bytes.len() {
+            let end_index = min(self.all_bytes.len(), address + width);
+            address..end_index
+        }
+        else {
+            0..0
+        }
+    }
+
+
+    /// Print `self.width` bytes from `address` or cut off if at end of all bytes
+    pub fn bytes_from(&self, address:usize) -> &[u8] {
+        &self.all_bytes[self.bytes_range_from(address)]
+    }
+
+
+    pub fn bytes_line(&self, bytes:&[u8], line_number:usize) -> String {
+        bytes_line_bytes(bytes, line_number, self.width).iter().map(|x| formatted_byte(*x, self.color)).collect::<Vec<String>>().join(" ")
+    }
+
+
+    // TODO Do this padding stuff format!  Unclear why previous attempts
+    // have failed.
+    pub fn bytes_line_padding(&self, bytes:&[u8], line_num:usize) -> String{
+        let mut to_return = String::new();
+        let expected_length = usize::from(self.width) * 3 - 1;
+        for _ in num_graphemes(&self.bytes_line(bytes, line_num))..expected_length {
+            to_return += " ";
+        }
+
+        to_return
+    }
+
+
+    /// returns index of the byte in the 0-th column of the main row printed
+    /// (not the context rows)
+    pub fn print_bytes(&self) -> Option<usize> {
+        if self.empty() {
+            return None;
+        }
+
+        let max = self.max_index();
+        if max.is_err() {
+            println!("? ({:?})", max);
+            return None;
+        }
+        let max = max.unwrap();
+
+        // TODO do this more prettily
+        let from = self.index.saturating_sub(self.before_context * usize::from(self.width));
+
+        let to = min(max, self.index + (self.after_context + 1) * usize::from(self.width) - 1);
+
+        /* Call the more specific function */
+        self.print_bytes_sans_context((from, to));
+
+        return Some(min(max, self.index + usize::from(self.width)));
+    }
+
     /// returns index of the byte in the 0-th column of the last row printed
-    pub fn print_bytes(&self, range:(usize, usize)) -> Option<usize> {
+    /// This is more primitive than `print_bytes`.  It just prints the bytes
+    /// from the range.
+    pub fn print_bytes_sans_context(&self, range:(usize, usize)) ->
+            Option<usize> {
         if self.empty() {
             return None;
         }
@@ -186,8 +261,8 @@ impl State {
         let from = range.0;
         let to = min(max, range.1);
         if bad_range(&self.all_bytes, (from, to)) {
-        println!("? (Bad range: ({}, {}))", range.0, range.1);
-        return None;
+            println!("? (Bad range: ({}, {}))", range.0, range.1);
+            return None;
         }
 
         let bytes = &self.all_bytes[from..=to];
@@ -200,8 +275,8 @@ impl State {
 
             print!(
                 "{}{}",
-                bytes_line(bytes, bytes_line_num, self.width, self.color),
-                bytes_line_padding(bytes, bytes_line_num, self.width)
+                self.bytes_line(bytes, bytes_line_num),
+                self.bytes_line_padding(bytes, bytes_line_num)
             );
 
             if self.show_chars {
@@ -251,6 +326,17 @@ impl fmt::Display for State {
                 string_from_radix(self.radix));
         to_write += &format!("Printing a newline every {} bytes\n",
                 hex_unless_dec_with_radix(usize::from(self.width), self.radix));
+
+        if self.before_context > 0 {
+            to_write += &format!("Printing {} lines before current line\n",
+            hex_unless_dec_with_radix(self.before_context, self.radix));
+        };
+
+        if self.after_context > 0 {
+            to_write += &format!("Printing {} lines after current line\n",
+            hex_unless_dec_with_radix(self.after_context, self.radix));
+        };
+
         if self.unsaved_changes {
             to_write += &format!("Unwritten changes")
         }
@@ -382,6 +468,54 @@ pub fn max_bytes_line(bytes:&[u8], width:NonZeroUsize) -> usize {
 mod tests {
     use super::*;
 
+    #[test]
+    fn test_bytes_from() {
+        let hex_twelve = [
+                0x00, 0x01, 0x02,
+                0x03, 0x04, 0x05,
+                0x06, 0x07, 0x08,
+                0x09, 0x0a, 0x0b,
+                0x0c, 0x0d, 0x0e,
+                0x0f, 0x10, 0x11,
+                0x12,
+            ];
+        let mut state = State {
+            radix: 16,
+            show_byte_numbers: true,
+            show_chars: true,
+            unsaved_changes: true,
+            last_search: None,
+            filename: "filename".to_owned(),
+            show_prompt: true,
+            color: true,
+            index: 0,
+            // height: NonZeroUsize::new(1).unwrap(),
+            width: NonZeroUsize::new(0x10).unwrap(),
+            all_bytes: Vec::from(hex_twelve),
+            n_padding: "   ".to_owned(),
+        };
+
+        assert_eq!(state.bytes_from(0), &hex_twelve[0x00..=0x0f]);
+        assert_eq!(state.bytes_from(1), &hex_twelve[0x01..=0x10]);
+        assert_eq!(state.bytes_from(2), &hex_twelve[0x02..=0x11]);
+        assert_eq!(state.bytes_from(3), &hex_twelve[0x03..=0x12]);
+        assert_eq!(state.bytes_from(4), &hex_twelve[0x04..=0x12]);
+        assert_eq!(state.bytes_from(5), &hex_twelve[0x05..=0x12]);
+        assert_eq!(state.bytes_from(6), &hex_twelve[0x06..=0x12]);
+        assert_eq!(state.bytes_from(7), &hex_twelve[0x07..=0x12]);
+        assert_eq!(state.bytes_from(11), &hex_twelve[0x0b..=0x12]);
+        assert_eq!(state.bytes_from(17), &hex_twelve[0x11..=0x12]);
+        assert_eq!(state.bytes_from(18), &hex_twelve[0x12..=0x12]);
+        assert_eq!(state.bytes_from(19), &hex_twelve[0..0]);
+        state.width = NonZeroUsize::new(3).unwrap();
+        assert_eq!(state.bytes_from(0), &hex_twelve[0x00..=0x02]);
+        assert_eq!(state.bytes_from(1), &hex_twelve[0x01..=0x03]);
+        assert_eq!(state.bytes_from(11), &hex_twelve[0x0b..=0x0d]);
+        assert_eq!(state.bytes_from(17), &hex_twelve[0x11..=0x12]);
+        assert_eq!(state.bytes_from(18), &hex_twelve[0x12..=0x12]);
+        assert_eq!(state.bytes_from(19), &hex_twelve[0..0]);
+    }
+
     #[test]
     fn test_byte_numbers() {
         let mut state = State {
@@ -389,6 +523,7 @@ mod tests {
             show_byte_numbers: true,
             show_chars: true,
             unsaved_changes: true,
+            last_search: None,
             filename: "filename".to_owned(),
             show_prompt: true,
             color: true,
@@ -578,23 +713,6 @@ pub fn bytes_line_bytes(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> &
 }
 
 
-// TODO Do this padding stuff format!  Unclear why previous attempts
-// have failed.
-pub fn bytes_line_padding(bytes:&[u8], line_num:usize, width:NonZeroUsize) -> String{
-    let mut to_return = String::new();
-    let expected_length = usize::from(width) * 3 - 1;
-    for _ in num_graphemes(&bytes_line(bytes, line_num, width, false))..expected_length {
-        to_return += " ";
-    }
-
-    to_return
-}
-
-pub fn bytes_line(bytes:&[u8], line_number:usize, width:NonZeroUsize, color:bool) -> String {
-    bytes_line_bytes(bytes, line_number, width).iter().map(|x| formatted_byte(*x, color)).collect::<Vec<String>>().join(" ")
-}
-
-
 pub fn chars_line_chars(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> Vec<char> {
     let mut to_return:Vec<char> = vec![];
     for index in bytes_line_range(bytes, line_number, width) {
